/*	$ DOOROOS : adc.c, v 0.1 2014/07/18 $	*/

/*
 * Copyright (C) 2014	Jung Byoung-Oh ibosn systems
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are not permitted completely.
 *
 */

/*
 * Module Name: adc.c
 *
 * Purpose: adc interface for dooroos
 */

#include <types.h>
#include <thread.h>
#include <memory.h>
#include <message.h>
#include <interrupt.h>
#include <device.h>
#include <error.h>

#include <window.h>

#include <mach/hal_interrupt.h>

#include <mach/SAMA5D3X.h>
#include <mach/SAMA5D3Xbd.h>

#include <mach/gpio.h>

#include <hal_panic.h>

#include <dev/drvlib.h>
#include <dev/adcdrv.h>
//#include <mach/pmc.h>
//#include <mach/tc.h>

#include <cache.h>

#include "adc.h"


#define DBG_LVL_ADC		0


#define MAX_ADC_DEV         1
#define DMA_ADDR_OFFSET         0x20000000
#define DMA_BUFFER_SIZE         (MEMPOOL_SIZE*2)    // 8K


#define TIMER_ENABLE
typedef struct _ADC_SERVER_ 
{
    AT91PS_USART        pUSARTReg;
    UINT32              nDeviceID;

   
    UINT16              wValue[NUM_CHANNELS];
    UINT32              wDone;

    UINT32              dwAdcClock;
    UINT32              EnableChannels;

	PBYTE				pTailBuf;
	UINT32				nTailBuf;

#ifdef TIMER_ENABLE

#define ADC_TIMER_DELAY          1000
#define ADC_TIMER_BITMAP         0x100
    HANDLE       hAdcDrvTimer;

#endif

    HANDLE  hThreadWait;
    VOID *  pBufferWait;

}ADC_SERVER;


STATIC ADC_SERVER        AdcServerData[MAX_ADC_DEV];

//STATIC HDMA_TRANSFER_DESC  TranDesc[RX_BUFFERS];

// Configure PIO controllers to periph mode
static struct pio_desc ADpin_pio[ADC_CHANNEL_MAX];
static const struct pio_desc ADpin_All[] = {
    {"AD0",     AT91C_PIN_PD(20),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD1",     AT91C_PIN_PD(21),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD2",     AT91C_PIN_PD(22),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD3",     AT91C_PIN_PD(23),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD4",     AT91C_PIN_PD(24),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD5",     AT91C_PIN_PD(25),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD6",     AT91C_PIN_PD(26),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD7",     AT91C_PIN_PD(27),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD8",     AT91C_PIN_PD(28),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD9",     AT91C_PIN_PD(29),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD10",    AT91C_PIN_PD(30),   0, PIO_DEFAULT, PIO_PERIPH_A},
    {"AD11",    AT91C_PIN_PD(31),   0, PIO_DEFAULT, PIO_PERIPH_A},
};

// Configure PIO controllers to periph mode
static const struct pio_desc ADTRG_pio[] = {
    {"ADTRG",     AT91C_PIN_PD(19),   0, PIO_DEFAULT, PIO_PERIPH_A},
};


//--------------------------------------------------------------
// Defines
//--------------------------------------------------------------
/*
 * We use two ADC channel for this example:
 *    ADC_CHANNEL_5  (potentiometer)
 *    ADC_CHANNEL_15 (temperature sensor)
 */

/** ADC convention done mask */
#define ADC_DONE_MASK   ( (1<<NUM_CHANNELS) - 1 )

/** Size of the receive buffer and transmit buffer. */
#define BUFFER_SIZE     NUM_CHANNELS

/** Reference voltage for ADC, in mv */
#define VOLT_REF        (3300)

/** The maximal digital value */
#define MAX_DIGITAL     (4095)






// Sample filters
#define MAX_DIFFERENCE_CHANGE		100
#define SAMPLING_TIME				0x3FFF

//DWORD gIntrTouch        = 0;
//STATIC UINT32 ADCClock = 0;

#define ABS(a)              ((a)>=0 ? (a): -(a))

//===========================================================================


STATIC INT32 Adc_Control_Mode( ADC_SERVER *pAdcServer, ADC_MODE_CONTROL *pAdcModeControl )
{
    INT32   ret = 0;

    pAdcServer->EnableChannels = pAdcModeControl->EnableChannels;

    if( pAdcServer->EnableChannels == 0 )
        return 0;

    // make pio setup
    INT32   iChannel;
    INT32   tChannel = 0;
    for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
    {
        if( pAdcServer->EnableChannels & (1<<iChannel) )
        {
            ADpin_pio[tChannel] = ADpin_All[iChannel];
            tChannel++;
//DBGLINE(1,"iChannel =%x\n", iChannel );
        }
    }

    pio_setup(ADpin_pio, tChannel);

#if 0
    /* Enable/disable sequencer */
    if ( pAdcModeControl->bSequenceEn )
    {
        UINT32  Seq1 = 0;
        UINT32  Seq2 = 0;
        UINT32  iOffset = 0;

        for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
        {
            if( (pAdcServer->EnableChannels & (1<<iChannel)) == 0)
                continue;

            if( iOffset > 32)
            {
                Seq2 = ADC_CHANNEL_11 << (iOffset-32); 
                iOffset += 4;
            }
            else
            {
                Seq1 = ADC_CHANNEL_1 << iOffset;
                iOffset += 4;
            }
        }

        /* Set user defined channel sequence */
        ADC_SetSequence( AT91C_BASE_TSADC, Seq1, Seq2 ) ;
        /* Enable sequencer */
        ADC_SetSequenceMode(AT91C_BASE_TSADC, 1 ) ;

        /* Enable channels */
        for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
        {
            if( (pAdcServer->EnableChannels & (1<<iChannel)) == 0)
                continue;

            ADC_EnableChannel( AT91C_BASE_TSADC, iChannel ) ;
        }
        /* Update channel number */
        //pAdcServer->ucChNum[0] = ADC_CHANNEL_11 ;
        //pAdcServer->ucChNum[1] = ADC_CHANNEL_2 ;
        //pAdcServer->ucChNum[2] = ADC_CHANNEL_0 ;
    }
    else
#endif
    {
        /* Disable sequencer */
        ADC_SetSequenceMode( AT91C_BASE_TSADC, 0 ) ;

        /* Enable  channels */
        for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
        {
            if( (pAdcServer->EnableChannels & (1<<iChannel)) == 0)
                continue;

            ADC_EnableChannel( AT91C_BASE_TSADC, iChannel ) ;
        }

        /* Update channel number */
        //pAdcServer->ucChNum[0] = ADC_CHANNEL_0 ;
        //pAdcServer->ucChNum[1] = ADC_CHANNEL_2 ;
        //pAdcServer->ucChNum[2] = ADC_CHANNEL_11 ;
    }


    /* Set gain and offset (only single ended mode used here) */
    ADC_SetAnalogChange( AT91C_BASE_TSADC, 0 ) ; /* Disable analog change */
    if ( pAdcModeControl->bGainEn  )
    {
        ADC_SetAnalogChange( AT91C_BASE_TSADC, 1) ;
        ADC_SetChannelGain( AT91C_BASE_TSADC, ADC_CGR_GAIN5(2) ) ; /* gain = 2 */
    }
    else
    {
        ADC_SetChannelGain( AT91C_BASE_TSADC, 0 ) ; /* gain = 1 */
    }

    if ( pAdcModeControl->bOffsetEn )
    {
        ADC_SetAnalogChange( AT91C_BASE_TSADC, 1 ) ;
        ADC_SetChannelOffset( AT91C_BASE_TSADC, ADC_COR_OFF5 ) ;
    }
    else
    {
        ADC_SetChannelOffset( AT91C_BASE_TSADC, 0 ) ;
    }
    /* Set Auto Calibration Mode*/
    if(pAdcModeControl->bAutoCalibEn)
    {
        ADC_SetCalibMode(AT91C_BASE_TSADC);
        while(1)
        {
            if((ADC_GetStatus( AT91C_BASE_TSADC ) & ADC_ISR_EOCAL) == ADC_ISR_EOCAL)    break;
        }
    }
    /* Set power save */
    if ( pAdcModeControl->bPowerSaveEn )
    {
        ADC_SetSleepMode( AT91C_BASE_TSADC, 1 ) ;
    }
    else
    {
        ADC_SetSleepMode( AT91C_BASE_TSADC, 0 ) ;
    }

#if 0
    /* Transfer with/without DMA */
    if ( pAdcModeControl->bDmaEn )
    {
//        _AdcDma( dmaRxChannel, (void*)gAdcSampleData.wValue, NUM_CHANNELS );
        DMA_Start(pAdcServer);
    }
    else
    {
        //last channel;
        UINT32  LastChannel = 0;
        
        /* Enable Data ready interrupt */
        for( iChannel = 0; iChannel < ADC_CHANNEL_MAX; iChannel++ )
        {
            if( (pAdcServer->EnableChannels & (1<<iChannel)) == 0)
                continue;

            LastChannel = iChannel;
        }
        //ADC_EnableIt( AT91C_BASE_TSADC, 1<<LastChannel ) ;
    }

    /* Enable ADC interrupt */
    //IRQ_ConfigureIT( ID_ADC, 0, ADC_IrqHandler );
    //IRQ_EnableIT( ID_ADC );

    /* Configure trigger mode and start convention */
    switch ( pAdcModeControl->TriggerMode )
    {
        case TRIGGER_MODE_SOFTWARE:
            ADC_SetTrigger( AT91C_BASE_TSADC, 0 ) ; /* Disable hardware trigger */
            ADC_SetTriggerMode (AT91C_BASE_TSADC, ADC_TRGR_TRGMOD_NO_TRIGGER); /* No trigger, only software trigger can start conversions */
            break ;

        case TRIGGER_MODE_ADTRG:
            //PIO_Configure( &gPinADTRG, 1 ) ;
            //PIO_Configure( &pinTrg, 1 ) ;
            pio_setup(ADTRG_pio, sizeof(ADTRG_pio)/sizeof(struct pio_desc));
            ADC_SetTrigger( AT91C_BASE_TSADC, ADC_MR_TRGSEL_ADC_TRIG0 ) ;
            ADC_SetTriggerMode (AT91C_BASE_TSADC, ADC_TRGR_TRGMOD_EXT_TRIG_ANY); /* External Trigger Any Edge */
            break ;
#if 0
        case TRIGGER_MODE_TIMER :
            //DrvTimerStart(pAdcServer->hAdcDrvTimer);
            _ConfigureTimerTrigger() ;
            break ;
#endif
        case TRIGGER_MODE_PERIOD:
        {
            ADC_SetTrigger( AT91C_BASE_TSADC, 0 ) ; /* Disable hardware trigger */
            ADC_SetTriggerPeriod(AT91C_BASE_TSADC, pAdcModeControl->Period /*period*/);
            ADC_SetTriggerMode (AT91C_BASE_TSADC, ADC_TRGR_TRGMOD_PERIOD_TRIG); /* periodic  */
            break;
        }

        default :
            break ;
    }
#endif

    return ret;
}



STATIC VOID
Adc_Init( ADC_SERVER *pAdcServer )
{
//    UINT32 dw ;
    UINT32 dwAdcClock ;
    UINT32 dwMaxAdcClock ;

    //DEBUGMSG(1, "%s (%d): ========= Adc_Init\n", __FUNCTION__, __LINE__);


    pAdcServer->EnableChannels = 0;

    /* Stop */
//    DMAD_StopTransfer( &dmad, dmaRxChannel );
    //ADC_DisableIt( AT91C_BASE_TSADC, 0xFFFFFFFF );

    /* Initialize ADC */
    //ADC_Initialize( AT91C_BASE_TSADC, SAMA5D3_ID_ADC );
//    memset( (void*)&gAdcSampleData, 0, sizeof( gAdcSampleData ) );
//    CP15_coherent_dcache_for_dma((UINT32)&gAdcSampleData, ((UINT32)(&gAdcSampleData) + sizeof( gAdcSampleData )));
    /*
     * Formula: ADCClock = MCK / ( (PRESCAL+1) * 2 )
     * For example, MCK = 64MHZ, PRESCAL = 4, then:
     *     ADCClock = 64 / ((4+1) * 2) = 6.4MHz;
     */
    /* Set ADC clock */
    dwMaxAdcClock = PMC_SetPeriMaxClock(SAMA5D3_ID_ADC, BOARD_MCK);

    // Enable peripheral clock
    PMC_EnablePeripheral( SAMA5D3_ID_ADC ) ;

    dwAdcClock = ADC_SetClock( AT91C_BASE_TSADC, 0, dwMaxAdcClock ) ;
    //kprintf( "Set ADC clock to %d Hz\n\r", (int)dwAdcClock ) ;


    pAdcServer->dwAdcClock = dwAdcClock;
    /* Formula:
     *     Startup  Time = startup value / ADCClock
     *     Transfer Time = (TRANSFER * 2 + 3) / ADCClock
     *     Tracking Time = (TRACKTIM + 1) / ADCClock
     *     Settling Time = settling value / ADCClock
     * For example, ADC clock = 6MHz (166.7 ns)
     *     Startup time = 512 / 6MHz = 85.3 us
     *     Transfer Time = (1 * 2 + 3) / 6MHz = 833.3 ns
     *     Tracking Time = (0 + 1) / 6MHz = 166.7 ns
     *     Settling Time = 3 / 6MHz = 500 ns
     */
    /* Set ADC timing */
    ADC_SetTiming( AT91C_BASE_TSADC, ADC_MR_STARTUP_SUT512, ADC_MR_TRACKTIM( 0 ), ADC_MR_SETTLING_AST17 ) ;

    /* Enable channel number tag */
    ADC_SetTagEnable( AT91C_BASE_TSADC, 1 ) ;




	return;
}

STATIC UINT32       Channel;

STATIC ADC_MODE_CONTROL DefaultAdcModeControl = 
{
    //TRIGGER_MODE_SOFTWARE, //TRIGGER_MODE   TriggerMode;
    TRIGGER_MODE_PERIOD, //TRIGGER_MODE   TriggerMode;
    FALSE,      //BOOL          bDmaEn;
    FALSE,      //BOOL          bSequenceEn;
    FALSE,      //BOOL          bGainEn;
    FALSE,      //BOOL          bOffsetEn;
    FALSE,      //BOOL          bPowerSaveEn;
    FALSE,      //BOOL          bAutoCalibEn;
    312500,
    //212000,    // UINT32       Period in ns
                //UINT32        EnableChannels
        0
//    | (1<<ADC_CHANNEL_0)
//    | (1<<ADC_CHANNEL_1)
//    | (1<<ADC_CHANNEL_2) 
//    | (1<<ADC_CHANNEL_3) 
//    | (1<<ADC_CHANNEL_4) 
//    | (1<<ADC_CHANNEL_5) 
//    | (1<<ADC_CHANNEL_6) 
//    | (1<<ADC_CHANNEL_7) 
//    | (1<<ADC_CHANNEL_8) 
//    | (1<<ADC_CHANNEL_9) 
//    | (1<<ADC_CHANNEL_10)
//      (1<<ADC_CHANNEL_11)
};


STATIC INT32
th_AdcDevice( UINT32 Arg )
{
    JB_RECVMESSAGE      Message;
    HANDLE              hPort;
    HANDLE              Handle = NULL;
    PDEVICECONFIG       pDeviceConfig;

    ADC_SERVER*         pAdcServer;

    pDeviceConfig = (PDEVICECONFIG)Arg;
    
    hPort = PortGet( );    
    PortSetType( hPort, PORTTYPE_MESSAGE| PORTTYPE_INTERRUPT| PORTTYPE_SIGNAL );

    Channel = ChannelGet( hPort );

    pAdcServer = AdcServerData + pDeviceConfig->Minor;

    pDeviceConfig->Name[3] = (UINT32)pDeviceConfig->Minor + '1';
    pDeviceConfig->Channel = Channel;
    DeviceRegister( (PDEVICECONFIG)pDeviceConfig );

    Adc_Init( pAdcServer );

    /* set default */
    Adc_Control_Mode(pAdcServer, &DefaultAdcModeControl);

#ifdef TIMER_ENABLE
    pAdcServer->hAdcDrvTimer = DrvTimerCreate( Channel, ADC_TIMER_DELAY, ADC_TIMER_BITMAP );
    // start the timer
    if( pAdcServer->hAdcDrvTimer != NULL )
        DrvTimerStart(pAdcServer->hAdcDrvTimer); 
#endif

	while(1)
	{
		MessageReceive( hPort, &Message);

		switch(Message.Type)
		{
		case MESSAGE_TYPE_INTERRUPT:
		{
            continue;
        }

		case MESSAGE_TYPE_SIGNAL:
		{

#ifdef TIMER_ENABLE		
            if( Message.Param & ADC_TIMER_BITMAP )
            {
            AT91PS_TSADC pADC = (AT91PS_TSADC)AT91C_BASE_TSADC;
			//ADC_SetTriggerMode(pADC, ADC_TRGR_TRGMOD_PERIOD_TRIG);
			ADC_SetTriggerMode(pADC, ADC_TRGR_TRGMOD_PERIOD_TRIG);
            Sleep(30);
            if( pADC->ADC_IMR & ADC_IMR_PEN )
            {
    			ADC_SetTriggerMode(pADC, ADC_TRGR_TRGMOD_PEN_TRIG);
            }

                int i;
                for( i = 0; i < ADC_CHANNEL_MAX; i++ )
                {
                    if( pAdcServer->EnableChannels & (1<<i) )
                    {
                        pAdcServer->wValue[i] = pADC->ADC_CDR[i];
                        pAdcServer->wDone |= (1<<i) ;
//DBGLINE(1,": pAdcServer->wValue[%d] = 0x%x\n",i,pAdcServer->wValue[i]);
                    } 
                }

//            DBGLINE(1,": pADC->ADC_CDR[7] = %x\n", pADC->ADC_CDR[7]);
//            DBGLINE(1,": pADC->ADC_TSMR = %x\n", pADC->ADC_TSMR);
//            DBGLINE(1,": pADC->ADC_ISR = %x\n", pADC->ADC_ISR);
//            DBGLINE(1,": pADC->ADC_IMR = %x\n", pADC->ADC_IMR);
            }
#endif
            continue;
		}
		
		case MESSAGE_TYPE_DEVICE_READ:
        {
		    Message.Type = 0;
			break;
	    }

		case MESSAGE_TYPE_DEVICE_WRITE:
	    {
            Message.Type = 0;
            break;
	    }

		case MESSAGE_TYPE_DEVICE_CONTROL:
	    {
            UINT32      Ret = 0;
            PIOCONTROL_PARAM    pIoControlParam;
            if( Handle == NULL || Message.Handle != Handle )
            {
                ThreadSetLastError( Message.hSender, ERR_INVALID_HANDLE );
                Message.Type = FALSE;
                break;
            }

            pIoControlParam = (PIOCONTROL_PARAM)Message.Param;
		    switch( pIoControlParam->Cmd )
		    {
		    case IOCTRL_ADC_MODE:
                Adc_Control_Mode( pAdcServer, (ADC_MODE_CONTROL *)pIoControlParam->pInBuf );
		    	//PIO_Power_SetState(POWER_MAINBOARD_5V, bLedOn?FALSE:TRUE);
				//DBGLINE(1,"%s (%d) :  IOCTRL_POWER_5V = %d\n", __FUNCTION__, __LINE__,bLedOn);
                Ret = TRUE;
				break;
		    case IOCTRL_ADC_BUFFER:
                //DBGLINE(1,"%s (%d) :  IOCTRL_POWER_9V = %d\n", __FUNCTION__, __LINE__,~bLedOn);
		    	//PIO_Power_SetState(POWER_MAINBOARD_9V, bLedOn?FALSE:TRUE);
                Ret = TRUE;
				break;
		    case IOCTRL_ADC_READ_SINGLE:
		    {
                UINT16 *pBuf = (UINT16 *)pIoControlParam->pOutBuf;
                if( pBuf == NULL )
                {
                    Message.Type = 0;
                    ThreadSetLastError( Message.hSender, ERR_INVALID_PARAMETER );
                    break;
                }

                memcpy( (char *)pBuf, (char *)pAdcServer->wValue, NUM_CHANNELS*2);
                break;
            }

		    case IOCTRL_ADC_READ_BUFFER:
		    {
                UINT16 *pBuf = (UINT16 *)pIoControlParam->pOutBuf;
                if( pBuf == NULL )
                {
                    Message.Type = 0;
                    ThreadSetLastError( Message.hSender, ERR_INVALID_PARAMETER );
                    break;
                }

                memcpy( (char *)pBuf, (char *)pAdcServer->wValue, NUM_CHANNELS*2);
                break;
            }

            default:
                ThreadSetLastError( Message.hSender, ERR_INVALID_PARAMETER );
				break;
			}

            Message.Type = Ret;
			break;
	    }

		case MESSAGE_TYPE_DEVICE_OPEN:
        {
			Handle = ChannelOpenHandleGetFrom(Channel);
			Message.Type = Handle;
            break;
        }

		case MESSAGE_TYPE_DEVICE_CLOSE:
        {
		    Handle = NULL;
		    Message.Type = TRUE;
			break;
	    }

        default:
            break;
		}

        MessageReply(&Message);

	}

	return TRUE;	
}



//==========================================================================
#define ADCDEV_STACK_SIZE       (4096)

STATIC DEVICECONFIG             AdcDeviceConfig ={"ADC", 0, -1, NULL, NULL };

STATIC THREADCONFIG             AdcDev_Thread =
                                            {
                                                (UINT32)&AdcDeviceConfig,
                                                th_AdcDevice,
                                                (PBYTE)NULL,
                                                ADCDEV_STACK_SIZE,
                                                THREAD_PRIORITY_DRIVER_HIGH+11
                                            };

//==========================================================================

#include <moduleloader.h>

PUBLIC HANDLE 
DllMain( HANDLE hDll, UINT32 Reason, PVOID Reserved )
{
    HANDLE hThread = NULL;
    switch( Reason )
    {
    case DLL_PROCESS_ATTACH:
        hThread = ThreadCreate (&AdcDev_Thread);
    	break;
    
    case DLL_PROCESS_DETACH:
    	break;
    }

    return hThread;
}

